######################################################################
#
# EPrint Session
#
#  Holds information about a particular EPrint session.
#
#
#  Fields are:
#    database        - EPrints::Database object
#    renderer        - EPrints::HTMLRender object
#
######################################################################
#
# License for eprints.org software version: Version 1.1.2 (29/08/2001)
# 
# Copyright (C) 2001, University of Southampton
# 
# The University of Southampton retains the copyright of this software
# code with the exception of the open archives component (in the
# openarchives/ directory), which is a modified version of code
# distributed by Cornell University Digital Library Research Group.
# 
# This software is freely distributable. Modified versions of this
# software may be distributed provided that a file README is included
# describing the modifications and from where the original version may
# be obtained.
# 
# This software is provided with no guarantees of suitability for any
# intended purpose. Use of the software is entirely at the end user's
# risk.
#
######################################################################

package EPrints::Session;

use Apache;

use EPrints::Database;
use EPrints::HTMLRender;

use EPrintSite::SiteRoutines;
use EPrintSite::SiteInfo;

use strict;


# GLOBAL SITE REVISION NUMBER
$EPrints::Session::eprints_software_version = "Version 1.1.2 (29/08/2001)";


######################################################################
#
# new( $offline )
#
#  Start a new EPrints session, opening a database connection,
#  creating a CGI query object and any other necessary session state
#  things.
#
#  Command line scripts should pass in true for $offline.
#  Apache-invoked scripts can omit it or pass in 0.
#
######################################################################

sub new
{
	my( $class, $offline ) = @_;
	
	my $self = {};
	bless $self, $class;
	
	# Get the Apache request object
	#$self->{request} = Apache::request( "POST" );

	# Create an HTML renderer object
	$self->{render} = EPrints::HTMLRender->new( $self, $offline );

	# Create a database connection
	$self->{database} = EPrints::Database->new();
	
	if( !defined $self->{database} )
	{
		# Database connection failure - noooo!
		$self->failure( "Failed to connect to database" );
	}

#$self->{starttime} = gmtime( time );

#EPrints::Log::debug( "Session", "Started session at $self->{starttime}" );
	
	EPrintSite::SiteRoutines::session_init( $self, $offline );

#
#	my @params = $self->{render}->{query}->param();
#	
#	foreach (@params)
#	{
#		my @vals = $self->{render}->{query}->param($_);
#		EPrints::Log::debug( "Session", "Param <$_> Values:<".@vals.">" );
#	}
	

	return( $self );
}


######################################################################
#
# failure()
#
#  Print an error messages describing why an operation has failed.
#
######################################################################

sub failure
{
	my( $self, $problem ) = @_;
	
	$self->{render}->render_error( $problem,
	                               $EPrintSite::SiteInfo::frontpage,
	                               $EPrintSite::SiteInfo::sitename );
}


######################################################################
#
# terminate()
#
#  Perform any cleaning up necessary
#
######################################################################

sub terminate
{
	my( $self ) = @_;
	
#EPrints::Log::debug( "Session", "Closing session started at $self->{starttime}" );
	EPrintSite::SiteRoutines::session_close( $self );

	$self->{database}->disconnect();

}


######################################################################
#
# mail_administrator( $subject, $message )
#
#  Sends a mail to the site administrator with the given subject and
#  message body.
#
######################################################################

sub mail_administrator
{
	my( $self, $subject, $message ) = @_;
	
	my $message_body = "Site maintenance message generated at ".gmtime( time );
	$message_body .= "\n\n$message\n";

	EPrints::Mailer::send_mail(
		"Site Administrator",
		$EPrintSite::SiteInfo::admin,
		$subject,
		$message_body );
}



1;
